//
// Benchmark for compiling a set of simple OpenCL kernels using a varying number of programs
// 
// Author: Karl Rupp,  me@karlrupp.net
// License: MIT/X11 license, see file LICENSE.txt
//


#include <iostream>
#include <string>
#include <vector>
#include <cmath>
#include <sstream>
#include <cassert>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include "benchmark-utils.hpp"

// The actual kernel, assigning 1 to the i-th element of a memory buffer
std::string generate_opencl_program(size_t program_id, size_t kernels_per_program)
{
  std::stringstream ss;

  for (std::size_t i=0; i<kernels_per_program; ++i)
    ss << "__kernel void kernel_" << program_id << "_" << i << " (__global float * x) { x[" << i << "] = 1; } " << std::endl;

  return ss.str();
}





int main(int argc, char **argv)
{
  typedef float       ScalarType;

  cl_int err;
  std::size_t num_kernels = 64;


  if (argc > 1)
    num_kernels = std::size_t(atoi(argv[1]));
  else
  {
    std::cout << "Usage: ./jit-overhead TOTAL_NUMBER_OF_KERNELS" << std::endl;
    exit(EXIT_FAILURE);
  }

  /////////////////////////// Part 1: Initialize OpenCL ///////////////////////////////////
    
  //
  // Query platform:
  //
  cl_uint num_platforms;
  cl_platform_id platform_ids[42];   //no more than 42 platforms supported...
  err = clGetPlatformIDs(42, platform_ids, &num_platforms); ERR_CHECK(err);

  std::cout << "# Platforms found: " << num_platforms << std::endl;
  for (cl_uint i=0; i<num_platforms; ++i)
  {
    char buffer[1024];
    cl_int err;
    err = clGetPlatformInfo(platform_ids[i], CL_PLATFORM_VENDOR, 1024 * sizeof(char), buffer, NULL); ERR_CHECK(err);
    
    std::stringstream ss;
    ss << "# (" << i << ") " << buffer << ": ";

    err = clGetPlatformInfo(platform_ids[i], CL_PLATFORM_VERSION, 1024 * sizeof(char), buffer, NULL); ERR_CHECK(err);

    ss << buffer;

    std::cout << ss.str() << std::endl;
  }

  std::size_t platform_index = 0;
  if (num_platforms > 1)
  {
    std::cout << "# Enter platform index to use: ";
    std::cin >> platform_index;
    platform_index = std::min<std::size_t>(platform_index, num_platforms - 1);
    std::cout << "#" << std::endl;
  }
  
  //
  // Query devices:
  //
  cl_device_id device_ids[42];
  cl_uint num_devices;
  err = clGetDeviceIDs(platform_ids[platform_index], CL_DEVICE_TYPE_ALL, 42, device_ids, &num_devices); ERR_CHECK(err);
  std::cout << "# Devices found: " << num_devices << std::endl;
  for (cl_uint i=0; i<num_devices; ++i)
  {
    char buffer[1024]; 
    cl_int err;          
    err = clGetDeviceInfo(device_ids[i], CL_DEVICE_NAME, sizeof(char)*1024, &buffer, NULL); ERR_CHECK(err);
    
    std::cout << "# (" << i << ") " << buffer << std::endl;
  }

  std::size_t device_index = 0;
  if (num_devices > 1)
  {
    std::cout << "# Enter index of device to use: ";
    std::cin >> device_index;
    device_index = std::min<std::size_t>(device_index, num_devices - 1);
    std::cout << "#" << std::endl;
  }

  // now set up a context containing the selected device:
  cl_context my_context = clCreateContext(0, 1, &(device_ids[device_index]), NULL, NULL, &err); ERR_CHECK(err);
   
  // create a command queue for the device:
  cl_command_queue queue = clCreateCommandQueue(my_context, device_ids[device_index], 0, &err); ERR_CHECK(err);


  /////////////////////////// Part 2: Run the 'autotuning' process ///////////////////////////////////

  std::cout << "# Running benchmark (this may take a while)..." << std::endl;
  std::cout << "# Programs, Kernels/Program, Compilation+Extraction, Labels" << std::endl;
  for (std::size_t num_progs = 1; num_progs <= num_kernels; num_progs *= 2)
  {
    std::vector<cl_program>  programs(num_progs);
    std::vector<std::string> program_sources(num_progs);
    std::vector<size_t>     program_sources_len(num_progs);
    std::size_t kernels_per_program = num_kernels / num_progs;
    
    for (std::size_t i=0; i<num_progs; ++i)
    {
      program_sources[i]     = generate_opencl_program(i, kernels_per_program);
      program_sources_len[i] = program_sources[i].length();
    }
    std::vector<cl_kernel> kernels(kernels_per_program);
    cl_uint num_kernels_ret = 0;

    Timer timer;
    timer.start();
    for (std::size_t i=0; i<num_progs; ++i)
    {
      const char * c_str = program_sources[i].c_str();
      programs[i] = clCreateProgramWithSource(my_context, 1, &c_str, &(program_sources_len[i]), &err); ERR_CHECK(err);
      err = clBuildProgram(programs[i], 0, NULL, NULL, NULL, NULL);
      if (err != CL_SUCCESS)
      {
        char buffer[8192];
        cl_build_status status;
        std::cout << "Build Scalar: Err = " << err;
        err = clGetProgramBuildInfo(programs[i], device_ids[device_index], CL_PROGRAM_BUILD_STATUS, sizeof(cl_build_status), &status, NULL); ERR_CHECK(err);
        err = clGetProgramBuildInfo(programs[i], device_ids[device_index], CL_PROGRAM_BUILD_LOG,    sizeof(char)*8192, &buffer, NULL); ERR_CHECK(err);
        std::cout << " Status = " << status << std::endl;
        std::cout << "Log: " << buffer << std::endl;
        std::cout << "Sources: " << program_sources[i] << std::endl;
      }
      err = clCreateKernelsInProgram(programs[i], kernels_per_program, &(kernels[0]), &num_kernels_ret); ERR_CHECK(err);
    }
    std::cout << num_progs << "    " << kernels_per_program << "     " << timer.get() << "      \"" << num_progs << "x" << kernels_per_program << "\"" << std::endl;

    assert(num_kernels_ret == kernels_per_program && bool("Number of kernels built invalid!"));

    // Clean up:
    for (std::size_t i = 0; i < kernels_per_program; ++i)
    {
      err = clReleaseKernel(kernels[i]); ERR_CHECK(err);
    }
    for (std::size_t i = 0; i < num_progs; ++i)
    {
      err = clReleaseProgram(programs[i]); ERR_CHECK(err);
    }
  }

  //
  // Part 3: Cleanup
  //
  err = clReleaseCommandQueue(queue); ERR_CHECK(err);
  err = clReleaseContext(my_context); ERR_CHECK(err);
  
  return EXIT_SUCCESS;
}

